@@ -47,6 +47,11 @@ const ccl_configuration default_config = {
  */
 #define GSL_EPSREL_SIGMAR 1E-5
 
+/**
+ * Relative precision in k_NL calculations
+ */
+#define GSL_EPSREL_KNL 1E-5
+
 /**
  * Relative precision in distance calculations
  */
@@ -70,6 +75,7 @@ const ccl_gsl_params default_gsl_params = {
   GSL_EPSREL,                          // INTEGRATION_LIMBER_EPSREL
   GSL_EPSREL_DIST,                     // INTEGRATION_DISTANCE_EPSREL
   GSL_EPSREL_SIGMAR,                   // INTEGRATION_SIGMAR_EPSREL
+  GSL_EPSREL_KNL,                      // INTEGRATION_KNL_EPSREL
   GSL_EPSREL,                          // ROOT_EPSREL
   GSL_N_ITERATION,                     // ROOT_N_ITERATION
   GSL_EPSREL_GROWTH,                   // ODE_GROWTH_EPSREL
@@ -86,6 +92,7 @@ const ccl_gsl_params default_gsl_params = {
 #undef GSL_N_ITERATION
 #undef GSL_INTEGRATION_GAUSS_KRONROD_POINTS
 #undef GSL_EPSREL_SIGMAR
+#undef GSL_EPSREL_KNL
 #undef GSL_EPSREL_DIST
 #undef GSL_EPSREL_GROWTH
 #undef GSL_EPSREL_DNDZ
@@ -98,17 +105,21 @@ const ccl_spline_params default_spline_params = {
   0.1,  // A_SPLINE_MIN
   0.01,  // A_SPLINE_MINLOG_PK
   0.1,  // A_SPLINE_MIN_PK,
+  0.01,  // A_SPLINE_MINLOG_SM,
+  0.1,  // A_SPLINE_MIN_SM,
   1.0,  // A_SPLINE_MAX,
   0.0001,  // A_SPLINE_MINLOG,
   250,  // A_SPLINE_NLOG,
 
   // mass splines
   0.025,  // LOGM_SPLINE_DELTA
-  440,  // LOGM_SPLINE_NM
+  50,  // LOGM_SPLINE_NM
   6,  // LOGM_SPLINE_MIN
   17,  // LOGM_SPLINE_MAX
 
   // PS a and k spline
+  13,  // A_SPLINE_NA_SM
+  6,  // A_SPLINE_NLOG_SM
   40,  // A_SPLINE_NA_PK
   11,  // A_SPLINE_NLOG_PK
 
@@ -117,6 +128,7 @@ const ccl_spline_params default_spline_params = {
   1E3,  // K_MAX
   5E-5,  // K_MIN
   0.025,  // DLOGK_INTEGRATION
+  5.,  // DCHI_INTEGRATION
   167,  // N_K
   100000,  // N_K_3DCOR
 
@@ -236,7 +248,6 @@ fgrowth: logarithmic derivative of the growth (density) (dlnD/da?)
 E: E(a)=H(a)/H0
 growth0: growth at z=0, defined to be 1
 sigma: ?
-p_lin: linear matter power spectrum at z=0?
 p_lnl: nonlinear matter power spectrum at z=0?
 computed_distances, computed_growth,
 computed_power, computed_sigma: store status of the computations
@@ -264,25 +275,22 @@ ccl_cosmology * ccl_cosmology_create(ccl_parameters params, ccl_configuration co
   cosmo->data.achi = NULL;
 
   cosmo->data.logsigma = NULL;
-  cosmo->data.dlnsigma_dlogm = NULL;
 
   cosmo->data.rsd_splines[0] = NULL;
   cosmo->data.rsd_splines[1] = NULL;
   cosmo->data.rsd_splines[2] = NULL;
 
-  cosmo->data.p_lin = NULL;
-  cosmo->data.p_nl = NULL;
   cosmo->computed_distances = false;
   cosmo->computed_growth = false;
-  cosmo->computed_linear_power = false;
-  cosmo->computed_nonlin_power = false;
   cosmo->computed_sigma = false;
   cosmo->status = 0;
   ccl_cosmology_set_status_message(cosmo, "");
 
   if(cosmo->spline_params.A_SPLINE_MAX !=1.) {
     cosmo->status = CCL_ERROR_SPLINE;
-    ccl_cosmology_set_status_message(cosmo, "ccl_core.c: A_SPLINE_MAX needs to be 1.\n");
+    ccl_cosmology_set_status_message(cosmo,
+                                     "ccl_core.c: ccl_cosmology_create(): "
+                                     "A_SPLINE_MAX needs to be 1.\n");
   }
 
   return cosmo;
@@ -372,17 +380,17 @@ A_s: amplitude of the primordial PS
 n_s: index of the primordial PS
 
  */
-ccl_parameters ccl_parameters_create(
-                     double Omega_c, double Omega_b, double Omega_k,
+ccl_parameters ccl_parameters_create(double Omega_c, double Omega_b, double Omega_k,
 				     double Neff, double* mnu, int n_mnu,
 				     double w0, double wa, double h, double norm_pk,
 				     double n_s, double bcm_log10Mc, double bcm_etab,
 				     double bcm_ks, double mu_0, double sigma_0,
+				     double c1_mg, double c2_mg, double lambda_mg,
 				     int nz_mgrowth, double *zarr_mgrowth,
 				     double *dfarr_mgrowth, int *status)
 {
   #ifndef USE_GSL_ERROR
-    gsl_set_error_handler_off ();
+    gsl_set_error_handler_off();
   #endif
 
   ccl_parameters params;
@@ -407,8 +415,8 @@ ccl_parameters ccl_parameters_create(
     params.N_nu_mass = 0;
   }else{
     params.N_nu_mass = n_mnu;
-   } 
-  
+   }
+
   // Dark Energy
   params.w0 = w0;
   params.wa = wa;
@@ -441,6 +449,9 @@ ccl_parameters ccl_parameters_create(
   // Params of the mu / Sigma parameterisation of MG
   params.mu_0 = mu_0;
   params.sigma_0 = sigma_0;
+  params.c1_mg = c1_mg;
+  params.c2_mg = c2_mg;
+  params.lambda_mg = lambda_mg;
 
   // Set remaining standard and easily derived parameters
   ccl_parameters_fill_initial(&params, status);
@@ -465,30 +476,6 @@ ccl_parameters ccl_parameters_create(
 }
 
 
-/* ------- ROUTINE: ccl_parameters_create_flat_lcdm --------
-INPUT: some cosmological parameters needed to create a flat LCDM model
-TASK: call ccl_parameters_create to produce an LCDM model
-*/
-ccl_parameters ccl_parameters_create_flat_lcdm(double Omega_c, double Omega_b, double h,
-                                               double norm_pk, double n_s, int *status)
-{
-  double Omega_k = 0.0;
-  double Neff = 3.046;
-  double w0 = -1.0;
-  double wa = 0.0;
-  double *mnu;
-  double mnuval = 0.;  // a pointer to the variable is not kept past the lifetime of this function
-  mnu = &mnuval;
-  double mu_0 = 0.;
-  double sigma_0 = 0.;
-
-  ccl_parameters params = ccl_parameters_create(Omega_c, Omega_b, Omega_k, Neff,
-						mnu, 0, w0, wa, h, norm_pk, n_s, -1, -1, -1, mu_0, sigma_0, -1, NULL, NULL, status);
-  return params;
-
-}
-
-
 /**
  * Write a cosmology parameters object to a file in yaml format.
  * @param cosmo Cosmological parameters
@@ -555,6 +542,9 @@ void ccl_parameters_write_yaml(ccl_parameters * params, const char * filename, i
   // Modified gravity parameters
   WRITE_DOUBLE(mu_0);
   WRITE_DOUBLE(sigma_0);
+  WRITE_DOUBLE(c1_mg);
+  WRITE_DOUBLE(c2_mg);
+  WRITE_DOUBLE(lambda_mg);
 
   // Derived parameters
   WRITE_DOUBLE(sigma8);
@@ -596,7 +586,9 @@ ccl_parameters ccl_parameters_read_yaml(const char * filename, int *status) {
   if (!f) {
     *status = CCL_ERROR_FILE_READ;
     ccl_parameters bad_params;
-    ccl_raise_warning(CCL_ERROR_FILE_READ, "ccl_core.c: Failed to read parameters from file.");
+    ccl_raise_warning(CCL_ERROR_FILE_READ,
+                      "ccl_core.c: ccl_parameters_read_yaml(): "
+                      "Failed to read parameters from file.");
 
     return bad_params;
   }
@@ -653,6 +645,9 @@ ccl_parameters ccl_parameters_read_yaml(const char * filename, int *status) {
   // Modified gravity parameters
   READ_DOUBLE(mu_0);
   READ_DOUBLE(sigma_0);
+  READ_DOUBLE(c1_mg);
+  READ_DOUBLE(c2_mg);
+  READ_DOUBLE(lambda_mg);
 
   // Derived parameters
   READ_DOUBLE(sigma8);
@@ -694,7 +689,8 @@ ccl_parameters ccl_parameters_read_yaml(const char * filename, int *status) {
   if (*status) {
     ccl_raise_warning(
       *status,
-      "ccl_core.c: Structure of YAML file incorrect: %s",
+      "ccl_core.c: ccl_parameters_read_yaml():"
+      "Structure of YAML file incorrect: %s",
       filename);
   }
 
@@ -712,7 +708,7 @@ ccl_parameters ccl_parameters_read_yaml(const char * filename, int *status) {
     Neff, mnu, N_nu_mass,
     w0, wa, h, norm_pk,
     n_s, bcm_log10Mc, bcm_etab,
-    bcm_ks, mu_0, sigma_0, nz_mgrowth, z_mgrowth,
+    bcm_ks, mu_0, sigma_0, c1_mg, c2_mg, lambda_mg, nz_mgrowth, z_mgrowth,
     df_mgrowth, status);
 
   if(z_mgrowth) free(z_mgrowth);
@@ -737,10 +733,7 @@ void ccl_data_free(ccl_data * data) {
   gsl_spline_free(data->fgrowth);
   gsl_spline_free(data->E);
   gsl_spline_free(data->achi);
-  gsl_spline_free(data->logsigma);
-  gsl_spline_free(data->dlnsigma_dlogm);
-  ccl_f2d_t_free(data->p_lin);
-  ccl_f2d_t_free(data->p_nl);
+  gsl_spline2d_free(data->logsigma);
   ccl_f1d_t_free(data->rsd_splines[0]);
   ccl_f1d_t_free(data->rsd_splines[1]);
   ccl_f1d_t_free(data->rsd_splines[2]);
